"""Build OSC messages for client applications."""

# ruff: noqa: PLR2004
from __future__ import annotations

from datetime import datetime
from typing import Any

from fastosc.message import osc_message
from fastosc.message.arg_value import (
    ARG_TYPE_ARRAY_START,
    ARG_TYPE_ARRAY_STOP,
    ARG_TYPE_BLOB,
    ARG_TYPE_DOUBLE,
    ARG_TYPE_FALSE,
    ARG_TYPE_FLOAT,
    ARG_TYPE_INT,
    ARG_TYPE_INT64,
    ARG_TYPE_MIDI,
    ARG_TYPE_NIL,
    ARG_TYPE_RGBA,
    ARG_TYPE_STRING,
    ARG_TYPE_TIMETAG,
    ARG_TYPE_TRUE,
    SUPPORTED_ARG_TYPES,
    ArgValue,
)

from .parsing import osc_types

# ruff: noqa: PLR0912,B904


class BuildError(Exception):
    """Error raised when an incomplete message is trying to be built."""


class OscMessageBuilder:
    """Builds arbitrary OscMessage instances."""

    def __init__(self, address: str | None = None) -> None:
        """Initialize a new builder for a message.

        Args:
          - address: The osc address to send this message to.
        """
        self._address = address
        self._args: list[tuple[str, ArgValue]] = []

    @property
    def address(self) -> str | None:
        """Returns the OSC address this message will be sent to."""
        return self._address

    @address.setter
    def address(self, value: str) -> None:
        """Sets the OSC address this message will be sent to."""
        self._address = value

    @property
    def args(self) -> list[tuple[str, ArgValue]]:
        """Returns the (type, value) arguments list of this message."""
        return self._args

    def _valid_type(self, arg_type: str) -> bool:
        if arg_type in SUPPORTED_ARG_TYPES:
            return True
        if isinstance(arg_type, (list, tuple)):
            return all(self._valid_type(sub_type) for sub_type in arg_type)
        return False

    def add_arg(self, arg_value: ArgValue, arg_type: str | None = None) -> None:
        """Add a typed argument to this message.

        Args:
          - arg_value: The corresponding value for the argument.
          - arg_type: A value in ARG_TYPE_* defined in this class,
                      if none then the type will be guessed.
        Raises:
          - ValueError: if the type is not supported.::wq
        """
        if arg_type and not self._valid_type(arg_type):
            raise ValueError(f"arg_type must be one of {SUPPORTED_ARG_TYPES}, or an array of valid types")
        if not arg_type:
            arg_type = self._get_arg_type(arg_value)
        if isinstance(arg_type, (list, tuple)):
            self._args.append((ARG_TYPE_ARRAY_START, None))
            for v, t in zip(arg_value, arg_type):  # type: ignore[var-annotated, arg-type]
                self.add_arg(v, t)  # type: ignore[var-annotated, arg-type]
            self._args.append((ARG_TYPE_ARRAY_STOP, None))
        else:
            self._args.append((arg_type, arg_value))

    # The return type here is actually Union[str, List[<self>]], however there
    # is no annotation for a recursive type like this.
    def _get_arg_type(self, arg_value: ArgValue) -> str:
        """Guess the type of a value.
        Args:
          - arg_value: The value to guess the type of.
        Raises:
          - ValueError: if the type is not supported.
        """
        if isinstance(arg_value, str):
            arg_type: str | Any = ARG_TYPE_STRING
        elif isinstance(arg_value, bytes):
            arg_type = ARG_TYPE_BLOB
        elif arg_value is True:
            arg_type = ARG_TYPE_TRUE
        elif arg_value is False:
            arg_type = ARG_TYPE_FALSE
        elif isinstance(arg_value, datetime):
            arg_type = ARG_TYPE_TIMETAG
        elif isinstance(arg_value, int):
            arg_type = ARG_TYPE_INT64 if arg_value.bit_length() > 32 else ARG_TYPE_INT
        elif isinstance(arg_value, float):
            arg_type = ARG_TYPE_FLOAT
        elif isinstance(arg_value, tuple) and len(arg_value) == 4:
            is_midi = True
            for i in arg_value:
                if not isinstance(i, int):  # for now assume midi if its a tuple of all ints
                    is_midi = False
            arg_type = ARG_TYPE_MIDI if is_midi else [self._get_arg_type(v) for v in arg_value]
        elif isinstance(arg_value, (list, tuple)):
            arg_type = [self._get_arg_type(v) for v in arg_value]
        elif arg_value is None:
            arg_type = ARG_TYPE_NIL
        else:
            raise ValueError("Infered arg_value type is not supported")
        return arg_type

    def build(self) -> osc_message.OscMessage:
        """Builds an OscMessage from the current state of this builder.

        Raises:
          - BuildError: if the message could not be build or if the address
                        was empty.

        Returns:
          - an osc_message.OscMessage instance.
        """
        if not self._address:
            raise BuildError("OSC addresses cannot be empty")
        dgram = b""
        try:
            # Write the address.
            dgram += osc_types.write_string(self._address)
            if not self._args:
                dgram += osc_types.write_string(",")
                return osc_message.OscMessage(dgram)

            # Write the parameters.
            arg_types = "".join([arg[0] for arg in self._args])  # type: ignore[misc]
            dgram += osc_types.write_string("," + arg_types)
            for arg_type, value in self._args:
                if arg_type == ARG_TYPE_STRING:
                    dgram += osc_types.write_string(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_INT:
                    dgram += osc_types.write_int(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_INT64:
                    dgram += osc_types.write_int64(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_FLOAT:
                    dgram += osc_types.write_float(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_TIMETAG:
                    dgram += osc_types.write_timetag(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_DOUBLE:
                    dgram += osc_types.write_double(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_BLOB:
                    dgram += osc_types.write_blob(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_RGBA:
                    dgram += osc_types.write_rgba(value)  # type: ignore[arg-type]
                elif arg_type == ARG_TYPE_MIDI:
                    dgram += osc_types.write_midi(value)  # type: ignore[arg-type]
                elif arg_type in (
                    ARG_TYPE_TRUE,
                    ARG_TYPE_FALSE,
                    ARG_TYPE_ARRAY_START,
                    ARG_TYPE_ARRAY_STOP,
                    ARG_TYPE_NIL,
                ):
                    continue
                else:
                    raise BuildError(f"Incorrect parameter type found {arg_type}")  # type: ignore[str-bytes-safe]

            return osc_message.OscMessage(dgram)
        except osc_types.BuildError as be:
            raise BuildError(f"Could not build the message: {be}")



from __future__ import annotations

from datetime import datetime
from typing import Any, List, Union

from fastosc.message.parsing import osc_types

ArgValue = Union[
    None,
    str,
    bytes,
    bool,
    int,
    float,
    tuple,
    datetime,
    osc_types.MidiPacket,
    List[
        Union[
            None,
            str,
            bytes,
            bool,
            int,
            float,
            tuple,
            datetime,
            osc_types.MidiPacket,
            List[
                Union[
                    None,
                    str,
                    bytes,
                    bool,
                    int,
                    float,
                    tuple,
                    datetime,
                    osc_types.MidiPacket,
                    List[
                        Union[
                            None,
                            str,
                            bytes,
                            bool,
                            int,
                            float,
                            tuple,
                            datetime,
                            osc_types.MidiPacket,
                            List[
                                Union[
                                    None,
                                    str,
                                    bytes,
                                    bool,
                                    int,
                                    float,
                                    tuple,
                                    datetime,
                                    osc_types.MidiPacket,
                                    List[
                                        Union[
                                            None,
                                            str,
                                            bytes,
                                            bool,
                                            int,
                                            float,
                                            tuple,
                                            datetime,
                                            osc_types.MidiPacket,
                                            List[
                                                Union[
                                                    None,
                                                    str,
                                                    bytes,
                                                    bool,
                                                    int,
                                                    float,
                                                    tuple,
                                                    datetime,
                                                    osc_types.MidiPacket,
                                                    List[
                                                        Union[
                                                            None,
                                                            str,
                                                            bytes,
                                                            bool,
                                                            int,
                                                            float,
                                                            tuple,
                                                            datetime,
                                                            osc_types.MidiPacket,
                                                            List[
                                                                Union[
                                                                    None,
                                                                    str,
                                                                    bytes,
                                                                    bool,
                                                                    int,
                                                                    float,
                                                                    tuple,
                                                                    datetime,
                                                                    osc_types.MidiPacket,
                                                                    List[
                                                                        Union[
                                                                            None,
                                                                            str,
                                                                            bytes,
                                                                            bool,
                                                                            int,
                                                                            float,
                                                                            tuple,
                                                                            datetime,
                                                                            osc_types.MidiPacket,
                                                                            List[
                                                                                Union[
                                                                                    None,
                                                                                    str,
                                                                                    bytes,
                                                                                    bool,
                                                                                    int,
                                                                                    float,
                                                                                    tuple,
                                                                                    datetime,
                                                                                    osc_types.MidiPacket,
                                                                                    List[
                                                                                        Union[
                                                                                            None,
                                                                                            str,
                                                                                            bytes,
                                                                                            bool,
                                                                                            int,
                                                                                            float,
                                                                                            tuple,
                                                                                            datetime,
                                                                                            osc_types.MidiPacket,
                                                                                            List[Any],
                                                                                        ]
                                                                                    ],
                                                                                ]
                                                                            ],
                                                                        ]
                                                                    ],
                                                                ]
                                                            ],
                                                        ]
                                                    ],
                                                ]
                                            ],
                                        ]
                                    ],
                                ]
                            ],
                        ]
                    ],
                ]
            ],
        ]
    ],
]

ARG_TYPE_FLOAT = "f"
ARG_TYPE_DOUBLE = "d"
ARG_TYPE_INT = "i"
ARG_TYPE_INT64 = "h"
ARG_TYPE_STRING = "s"
ARG_TYPE_BLOB = "b"
ARG_TYPE_RGBA = "r"
ARG_TYPE_MIDI = "m"
ARG_TYPE_TIMETAG = "t"
ARG_TYPE_TRUE = "T"
ARG_TYPE_FALSE = "F"
ARG_TYPE_NIL = "N"

ARG_TYPE_ARRAY_START = "["
ARG_TYPE_ARRAY_STOP = "]"

SUPPORTED_ARG_TYPES = (
    ARG_TYPE_FLOAT,
    ARG_TYPE_DOUBLE,
    ARG_TYPE_INT,
    ARG_TYPE_INT64,
    ARG_TYPE_TIMETAG,
    ARG_TYPE_BLOB,
    ARG_TYPE_STRING,
    ARG_TYPE_RGBA,
    ARG_TYPE_MIDI,
    ARG_TYPE_TRUE,
    ARG_TYPE_FALSE,
    ARG_TYPE_NIL,
)



from __future__ import annotations

import logging
from abc import ABC

from fastosc.dispatcher import Dispatcher
from fastosc.server.server_base import OSCServerBase


class OSCDispatcherServer(OSCServerBase, ABC):
    def __init__(self, dispatcher: Dispatcher, logger: logging.Logger, local_addr: tuple[str, int]) -> None:
        super().__init__(logger, local_addr)
        self._dispatcher = dispatcher
        self._dispatcher.set_server(self)



# ruff: noqa: FBT001,FBT002
from __future__ import annotations

import inspect
import logging
from typing import Any, Callable

from fastosc.dispatcher import Dispatcher
from fastosc.dispatcher.handler import HandlerInfo
from fastosc.message.arg_value import ArgValue
from fastosc.message.osc_message import OscMessage


def _format_response(arg_value: ArgValue) -> list[ArgValue]:
    if isinstance(arg_value, list):
        return arg_value
    return [arg_value]


def wrapper(
    address: str,
    prefix_: str,
    include_original_message: bool = False,
    include_remote_addr: bool = False,
    listen: bool = False,
) -> Callable[..., Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]]]:
    if not address.startswith("/"):
        address = f"/{address}"

    # todo maybe clean up this code a little bit as it was changed to infer parameter types
    # from handler function signature vs annotation
    def check_accepts(
        f: Callable[..., Any],
    ) -> Callable:
        sign = inspect.signature(f)
        shape = [sign.parameters[p].annotation for p in sign.parameters]
        shape = [eval(t) if isinstance(t, str) else t for t in shape]
        if len(shape) == 1:
            shape = []
        if len(shape) > 1:
            shape = shape[1:]
        orignal_shape = shape
        if not shape:
            shape = []
        if not isinstance(shape, list):
            shape = [shape]

        shape = [OSCRouter, *shape]
        if include_remote_addr:
            shape = [*shape, tuple]
        if include_original_message:
            shape = [*shape, OscMessage]  # todo: osc.Message here as well for validation

        arg_count = f.__code__.co_argcount - 1

        shape_len = len(shape) - 1

        if shape_len < arg_count:
            raise InvalidParameterValueException(
                f"too many arguments, expected {shape_len} " f"arguments but signature has {arg_count}.",
            )
        if shape_len > arg_count:
            raise InvalidParameterValueException(
                f"too few arguments, expected {shape_len} " f"arguments but signature has {arg_count}.",
            )

        # do we do someting with remote_addres?
        # maybe include it if part of parameters? remote_addr param - inject as final param if needed
        def new_f(self: OSCRouter, original_args: list[ArgValue], remote_address: tuple[str, int]) -> list[ArgValue]:
            for a, t in zip(original_args, shape[1:] if len(shape) > 1 else []):
                q = "'" if isinstance(a, str) else ""
                if not isinstance(a, t):
                    raise InvalidParameterValueException(
                        f"arg {q}{a}{q} of type <{type(a).__name__}> "  # type: ignore[str-bytes-safe]
                        f"does not match <{t.__name__}>",
                    )
            args = original_args
            if include_remote_addr:
                args = [*original_args, remote_address]  # type: ignore[list-item]
            if include_original_message:
                args = [*args, original_args]  # type: ignore[list-item]
            return _format_response(f(self, *args))

        if listen:
            new_f.listen = True  # type: ignore[attr-defined]
        new_f.osc_handler = True  # type: ignore[attr-defined]
        new_f.address = f"/{prefix_}{address}"  # type: ignore[attr-defined]
        new_f.raw_address = address  # type: ignore[attr-defined]
        new_f.__name__ = f.__name__  # type: ignore[attr-defined]
        new_f.handler_info = HandlerInfo(  # type: ignore[attr-defined]
            shape=orignal_shape,
            signature=sign,
            doc=inspect.getdoc(f),
            function_name=f.__name__,
        )
        return new_f

    return check_accepts


def osc_get(
    address: str,
    include_original_message: bool = False,
    include_remote_addr: bool = False,
    listen: bool = True,
) -> Callable[..., Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]]]:
    return wrapper(
        address,
        "get",
        include_original_message=include_original_message,
        include_remote_addr=include_remote_addr,
        listen=listen,
    )


def osc_set(
    address: str,
    include_original_message: bool = False,
    include_remote_addr: bool = False,
) -> Callable[..., Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]]]:
    return wrapper(
        address,
        "set",
        include_original_message=include_original_message,
        include_remote_addr=include_remote_addr,
    )


def _create_listener_key(address: str, remote_addr: tuple[str, int], args: list[ArgValue]) -> str:
    return f"{remote_addr}|{address}|{args}"


class OSCRouter:
    _listeners: dict[str, Callable]
    _routers: list[OSCRouter]

    def _add_router(self, *, router: OSCRouter) -> None:
        self._routers.append(router)

    def _add_handler(
        self,
        *,
        address: str,
        handler: Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]],
        handler_info: HandlerInfo | None = None,
    ) -> None:
        self._dispatcher.add_handler(address=f"{self._namespace}{address}", handler=handler, handler_info=handler_info)

    # class for overriding so an implementing class can wire up the listener as sees fit,
    # but this class will maintain a dict / cache with address/callback ids etc
    def _add_listener(self, *, address: str, listener: Callable, args: list[Any]) -> Callable[[], None] | None:
        # the method returns how to remove itself
        # this will already be validated for the appropriate address so it should be safe,
        # but not sure how we will transfer the types properly.
        # this must be implemented by the implementing class
        # the idea is that you get the args, like /track/device/parameter args= [tidx,didx,pidx] _song.tracks[0
        #  if address == "/tempo"
        #   tidx, didx, pidx = args
        #   param = _song.tracks[tidx].devices[didx].parameters[pidx].add_tempo_listener(listener)
        # if hasattr(param, f"add_{address}_listener"):
        " the listener class should do a call to the get of the param with the args" ""

    # class for overriding so an implementing class can remove up the listener as sees fit
    def _remove_listener(self, address: str, listener: Callable, args: list[Any]) -> None:
        # here, we can just look up the result from the saved listener method and invoke it?
        # self.listeners[x] = _add_listener()
        # if self._listeners.get(x)]:
        #   self._listeners()
        pass

    def _setup_listener(
        self,
        h: Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]],
        start: bool = False,
        stop: bool = False,
    ) -> tuple[str, Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]]]:
        if not start and not stop:
            raise InvalidParameterValueException("Need to set either stop or start for listener")
        prefix = "/start_listen" if start else "/stop_listen"
        address: str = h.address  # type: ignore[attr-defined]
        raw_address: str = h.raw_address  # type: ignore[attr-defined]

        def listener(args: list[ArgValue], remote_address: tuple[str, int]) -> None:
            key = _create_listener_key(address=address, remote_addr=remote_address, args=args)
            self._logger.info(key)
            if start:
                if key not in self._listeners:

                    def callback() -> None:
                        self._dispatcher.send(
                            address=f"{self._namespace}{address}",
                            remote_addr=remote_address,
                            params=h(args, remote_address),
                            include_base_address=True,
                        )

                    added_listener = self._add_listener(address=raw_address, listener=callback, args=args)
                    if added_listener:
                        self._listeners[key] = added_listener
                        logging.info(f"listener added for {address} and args {args} for client@{remote_address}")
                    else:
                        logging.info(f"start listen called for {address} and args {args}, but no listener was added.")
                else:
                    logging.info(
                        f"listener already existing for {address} and args {args} for client@{remote_address}, "
                        f"will not add unless sending a specific callback query param trailing (|Nil|QueryId",
                    )
            elif stop and key in self._listeners:
                self._listeners[key]()
                logging.info(f"listener stopped for {address} and args {args} for client@{remote_address}")

            # this will also send a message on every start / stop request
            self._dispatcher.send(
                address=f"{self._namespace}{address}",
                remote_addr=remote_address,
                params=h(args, remote_address),
                include_base_address=True,
            )

        return f"{prefix}{h.raw_address}", listener  # type: ignore[attr-defined,return-value]

    def _setup_handlers(self) -> None:
        handlers = [d[1] for d in inspect.getmembers(self) if not d[0].startswith("_") and inspect.ismethod(d[1])]
        for h in handlers:
            if hasattr(h, "osc_handler") and h.osc_handler and hasattr(h, "address") and h.address:
                handler_info: HandlerInfo | None = None
                if hasattr(h, "handler_info"):
                    handler_info = h.handler_info
                self._add_handler(
                    address=h.address,
                    handler=h,
                    handler_info=handler_info,
                )
                if hasattr(h, "listen") and h.listen and hasattr(h, "raw_address") and h.raw_address:
                    start_address, start_listen = self._setup_listener(h, start=True)
                    self._add_handler(
                        address=start_address,
                        handler=start_listen,
                        handler_info=handler_info,
                    )
                    stop_address, stop_listen = self._setup_listener(h, stop=True)
                    self._add_handler(
                        address=stop_address,
                        handler=stop_listen,
                        handler_info=handler_info,
                    )

    def clear_listeners(self) -> None:
        pass
        # for listener in self._listeners:
        # call listener()
        #     if self._song.tempo_has_listener(listener):
        #         self._song.remove_tempo_listener(listener)

    def __init__(self, *, dispatcher: Dispatcher, namespace: str) -> None:
        self._dispatcher = dispatcher
        if not namespace.startswith("/"):
            namespace = f"/{namespace}"
        self._namespace = namespace
        self._logger = self._dispatcher._logger
        self._listeners = {}
        self._routers = []
        self._setup_handlers()

    def _clear_listeners(self) -> None:
        for stop_listener in self._listeners.values():
            stop_listener()


class InvalidParameterValueException(Exception):
    pass



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class HandlerInputParam:
    name: str
    type: str


@dataclass
class HandlerDescription:
    address: str
    params: list[HandlerInputParam]
    result_type: str
    doc: str | None = None
    function_name: str | None = None



from __future__ import annotations

import logging
import traceback
from abc import ABC, abstractmethod

from fastosc.message.arg_value import ArgValue
from fastosc.message.convert import convert_message
from fastosc.message.osc_message_builder import BuildError


class OSCServerBase(ABC):
    def __init__(self, logger: logging.Logger, local_addr: tuple[str, int]) -> None:
        self._logger = logger
        self._local_addr = local_addr
        self._logger.info("Starting OSC server (local %s)", str(self._local_addr))

    @abstractmethod
    def _send_bytes(self, data: bytes, remote_addr: tuple[str, int]) -> None:
        pass

    def send(self, *, address: str, params: list[ArgValue], remote_addr: tuple[str, int]) -> None:
        try:
            self._send_bytes(data=convert_message(address=address, params=params), remote_addr=remote_addr)
        except BuildError:
            self._logger.error(f"OSC build error: {traceback.format_exc()}")



from __future__ import annotations

import logging
import re
from typing import Callable

from fastosc.dispatcher.handler import HandlerInfo
from fastosc.docs import HandlerDescription, HandlerInputParam
from fastosc.message.arg_value import ArgValue
from fastosc.message.osc_bundle import OscBundle
from fastosc.message.osc_message import OscMessage
from fastosc.server.server_base import OSCServerBase

MAX_LINE_LENGTH = 45


class Dispatcher:
    def __init__(self, *, logger: logging.Logger, base_address: str = "") -> None:
        self._server: OSCServerBase | None = None
        self._callbacks: dict[str, Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]]] = {}
        self._logger = logger
        if not base_address.startswith("/"):
            base_address = f"/{base_address}"
        self._base_address = base_address
        self._handler_docs: list[HandlerDescription] = []

    def set_server(self, server: OSCServerBase) -> None:
        self._server = server

    def add_handler(
        self,
        *,
        address: str,
        handler: Callable[[list[ArgValue], tuple[str, int]], list[ArgValue]],
        handler_info: HandlerInfo | None = None,
    ) -> None:
        """
        Add an OSC handler.

        Args:
            address: The OSC address string
            handler: A handler function, with signature:
                     params: Tuple[Any, ...]
            handler_info: metadata about the incoming message
        """
        if not address.startswith("/"):
            address = f"/{address}"
        address = f"{self._base_address}{address}"
        route_log_info = f"Added route {address}"
        self._callbacks[address] = handler
        if handler_info:
            s = handler_info.signature
            params = [
                HandlerInputParam(name=s.parameters[p].name, type=s.parameters[p].annotation)
                for p in s.parameters
                if s.parameters[p].name != "self"
            ]
            handler_desc = HandlerDescription(
                params=params,
                doc=handler_info.doc,
                result_type=s.return_annotation,
                address=address,
                function_name=handler_info.function_name,
            )
            self._handler_docs.append(handler_desc)
            whitespace = (MAX_LINE_LENGTH - len(address)) * " "
            return_type_str = ""
            if handler_desc.result_type.startswith("list"):
                return_type_str = handler_desc.result_type[4:-1]
            elif handler_desc.result_type.startswith("tuple"):
                return_type_str = handler_desc.result_type[6:-1]
            else:
                return_type_str = handler_desc.result_type
            route_log_info = (
                f"{route_log_info}{whitespace}"
                f"[{', '.join([f'{p.name}:{p.type}' for p in handler_desc.params])}] "
                f"-> [{return_type_str}]"
            )
        self._logger.info(route_log_info)

    def clear_handlers(self) -> None:
        """
        Remove all existing OSC handlers.
        """

        self._callbacks = {}

    def send(
        self,
        *,
        address: str,
        remote_addr: tuple[str, int],
        params: list[ArgValue],
        include_base_address: bool = False,
    ) -> None:
        if self._server:
            if include_base_address:
                address = f"{self._base_address}{address}"
            self._server.send(address=address, params=params, remote_addr=remote_addr)
        else:
            self._logger.error(f"Trying to send OSC message to remote address {remote_addr}, but not server is set up")

    def process_message(self, *, message: OscMessage, remote_addr: tuple[str, int]) -> None:
        if message.address in self._callbacks:
            callback = self._callbacks[message.address]
            rv = callback(message.params, remote_addr)
            if rv:
                assert isinstance(rv, list)
                if rv != [None]:
                    self.send(address=message.address, params=rv, remote_addr=remote_addr)
        elif "*" in message.address:
            regex = message.address.replace("*", "[^/]+")
            for callback_address, callback in self._callbacks.items():
                if re.match(regex, callback_address):
                    try:
                        rv = callback(message.params, remote_addr)
                    except ValueError:
                        # --------------------------------------------------------------------------------
                        # Don't throw errors for queries that require more arguments
                        # (e.g. /live/track/get/send with no args)
                        # --------------------------------------------------------------------------------
                        continue
                    except AttributeError:
                        # --------------------------------------------------------------------------------
                        # Don't throw errors when trying to create listeners for properties that can't
                        # be listened for (e.g. can_be_armed, is_foldable)
                        # --------------------------------------------------------------------------------
                        continue
                    if rv is not None:
                        assert isinstance(rv, list)
                        self.send(address=callback_address, params=rv, remote_addr=remote_addr)
        else:
            self._logger.error(f"Unknown OSC address: {message.address}")
            # todo: return the error to the socket that sent it

    def process_bundle(self, *, bundle: OscBundle, remote_addr: tuple[str, int]) -> None:
        for i in bundle:
            if OscBundle.dgram_is_bundle(i.dgram):
                self.process_bundle(bundle=i, remote_addr=remote_addr)
            else:
                self.process_message(message=i, remote_addr=remote_addr)



from __future__ import annotations

import inspect
from dataclasses import dataclass


@dataclass
class HandlerInfo:
    signature: inspect.Signature
    doc: str | None
    shape: type | list[type] | None = None
    function_name: str | None = None



"""Functions to get OSC types from datagrams and vice versa"""

# ruff: noqa: FBT001,FBT002,B904
from __future__ import annotations

import math
import struct
from datetime import datetime, timedelta
from typing import Tuple, cast

from fastosc.message.parsing import ntp

MidiPacket = Tuple[int, int, int, int]


class ParseError(Exception):
    """Base exception for when a datagram parsing error occurs."""


class BuildError(Exception):
    """Base exception for when a datagram building error occurs."""


# Constant for special ntp datagram sequences that represent an immediate time.
IMMEDIATELY = 0

# Datagram length in bytes for types that have a fixed size.
_INT_DGRAM_LEN = 4
_INT64_DGRAM_LEN = 8
_UINT64_DGRAM_LEN = 8
_FLOAT_DGRAM_LEN = 4
_DOUBLE_DGRAM_LEN = 8
_TIMETAG_DGRAM_LEN = 8
# Strings and blob dgram length is always a multiple of 4 bytes.
_STRING_DGRAM_PAD = 4
_BLOB_DGRAM_PAD = 4
_EMPTY_STR_DGRAM = b"\x00\x00\x00\x00"


def write_string(val: str) -> bytes:
    """Returns the OSC string equivalent of the given python string.

    Raises:
      - BuildError if the string could not be encoded.
    """
    try:
        dgram = val.encode("utf-8")  # Default, but better be explicit.
    except (UnicodeEncodeError, AttributeError) as e:
        raise BuildError(f"Incorrect string, could not encode {e}")
    diff = _STRING_DGRAM_PAD - (len(dgram) % _STRING_DGRAM_PAD)
    dgram += b"\x00" * diff
    return dgram


def get_string(dgram: bytes, start_index: int) -> tuple[str, int]:
    """Get a python string from the datagram, starting at pos start_index.

    According to the specifications, a string is:
    "A sequence of non-null ASCII characters followed by a null,
    followed by 0-3 additional null characters to make the total number
    of bits a multiple of 32".

    Args:
      dgram: A datagram packet.
      start_index: An index where the string starts in the datagram.

    Returns:
      A tuple containing the string and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    if start_index < 0:
        raise ParseError("start_index < 0")
    offset = 0
    try:
        if len(dgram) > start_index + _STRING_DGRAM_PAD and dgram[start_index + _STRING_DGRAM_PAD] == _EMPTY_STR_DGRAM:
            return "", start_index + _STRING_DGRAM_PAD
        while dgram[start_index + offset] != 0:
            offset += 1
        # Align to a byte word.
        if (offset) % _STRING_DGRAM_PAD == 0:
            offset += _STRING_DGRAM_PAD
        else:
            offset += -offset % _STRING_DGRAM_PAD
        # Python slices do not raise an IndexError past the last index,
        # do it ourselves.
        if offset > len(dgram[start_index:]):
            raise ParseError("Datagram is too short")
        data_str = dgram[start_index : start_index + offset]
        return data_str.replace(b"\x00", b"").decode("utf-8"), start_index + offset
    except IndexError as ie:
        raise ParseError(f"Could not parse datagram {ie}")
    except TypeError as te:
        raise ParseError(f"Could not parse datagram {te}")


def write_int(val: int) -> bytes:
    """Returns the datagram for the given integer parameter value

    Raises:
      - BuildError if the int could not be converted.
    """
    try:
        return struct.pack(">i", val)
    except struct.error as e:
        raise BuildError(f"Wrong argument value passed: {e}")


def get_int(dgram: bytes, start_index: int) -> tuple[int, int]:
    """Get a 32-bit big-endian two's complement integer from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the integer starts in the datagram.

    Returns:
      A tuple containing the integer and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _INT_DGRAM_LEN:
            raise ParseError("Datagram is too short")
        return (struct.unpack(">i", dgram[start_index : start_index + _INT_DGRAM_LEN])[0], start_index + _INT_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")


def write_int64(val: int) -> bytes:
    """Returns the datagram for the given 64-bit big-endian signed parameter value

    Raises:
      - BuildError if the int64 could not be converted.
    """
    try:
        return struct.pack(">q", val)
    except struct.error as e:
        raise BuildError(f"Wrong argument value passed: {e}")


def get_int64(dgram: bytes, start_index: int) -> tuple[int, int]:
    """Get a 64-bit big-endian signed integer from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the 64-bit integer starts in the datagram.

    Returns:
      A tuple containing the 64-bit integer and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _INT64_DGRAM_LEN:
            raise ParseError("Datagram is too short")
        return (
            struct.unpack(">q", dgram[start_index : start_index + _INT64_DGRAM_LEN])[0],
            start_index + _INT64_DGRAM_LEN,
        )
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")


def get_uint64(dgram: bytes, start_index: int) -> tuple[int, int]:
    """Get a 64-bit big-endian unsigned integer from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the integer starts in the datagram.

    Returns:
      A tuple containing the integer and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _UINT64_DGRAM_LEN:
            raise ParseError("Datagram is too short")
        return (
            struct.unpack(">Q", dgram[start_index : start_index + _UINT64_DGRAM_LEN])[0],
            start_index + _UINT64_DGRAM_LEN,
        )
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")


def write_timetag(ts: float | datetime) -> bytes:
    # do conversion
    if isinstance(ts, datetime):
        ts = ts.timestamp()
    if ts > 0:
        fract, secs = math.modf(ts)
        binary = struct.pack(">LL", secs, fract)
    else:
        binary = struct.pack(">LL", 0, 1)

    return binary


def get_timetag(dgram: bytes, start_index: int) -> tuple[tuple[datetime, int], int]:
    """Get a 64-bit OSC time tag from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the osc time tag starts in the datagram.

    Returns:
      A tuple containing the tuple of time of sending in utc as datetime and the
      fraction of the current second and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _TIMETAG_DGRAM_LEN:
            raise ParseError("Datagram is too short")

        timetag, _ = get_uint64(dgram, start_index)
        seconds, fraction = ntp.parse_timestamp(timetag)

        hours, seconds = seconds // 3600, seconds % 3600
        minutes, seconds = seconds // 60, seconds % 60

        utc = datetime.combine(ntp._NTP_EPOCH, datetime.min.time()) + timedelta(
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            microseconds=fraction,
        )

        return (utc, fraction), start_index + _TIMETAG_DGRAM_LEN
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")


def write_float(val: float) -> bytes:
    """Returns the datagram for the given float parameter value

    Raises:
      - BuildError if the float could not be converted.
    """
    try:
        return struct.pack(">f", val)
    except struct.error as e:
        raise BuildError(f"Wrong argument value passed: {e}")


def get_float(dgram: bytes, start_index: int) -> tuple[float, int]:
    """Get a 32-bit big-endian IEEE 754 floating point number from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the float starts in the datagram.

    Returns:
      A tuple containing the float and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _FLOAT_DGRAM_LEN:
            # Noticed that Reaktor doesn't send the last bunch of \x00 needed to make
            # the float representation complete in some cases, thus we pad here to
            # account for that.
            dgram = dgram + b"\x00" * (_FLOAT_DGRAM_LEN - len(dgram[start_index:]))
        return (
            struct.unpack(">f", dgram[start_index : start_index + _FLOAT_DGRAM_LEN])[0],
            start_index + _FLOAT_DGRAM_LEN,
        )
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")


def write_double(val: float) -> bytes:
    """Returns the datagram for the given double parameter value

    Raises:
      - BuildError if the double could not be converted.
    """
    try:
        return struct.pack(">d", val)
    except struct.error as e:
        raise BuildError(f"Wrong argument value passed: {e}")


def get_double(dgram: bytes, start_index: int) -> tuple[float, int]:
    """Get a 64-bit big-endian IEEE 754 floating point number from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the double starts in the datagram.

    Returns:
      A tuple containing the double and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _DOUBLE_DGRAM_LEN:
            raise ParseError("Datagram is too short")
        return (
            struct.unpack(">d", dgram[start_index : start_index + _DOUBLE_DGRAM_LEN])[0],
            start_index + _DOUBLE_DGRAM_LEN,
        )
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")


def get_blob(dgram: bytes, start_index: int) -> tuple[bytes, int]:
    """Get a blob from the datagram.

    According to the specifications, a blob is made of
    "an int32 size count, followed by that many 8-bit bytes of arbitrary
    binary data, followed by 0-3 additional zero bytes to make the total
    number of bits a multiple of 32".

    Args:
      dgram: A datagram packet.
      start_index: An index where the float starts in the datagram.

    Returns:
      A tuple containing the blob and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    size, int_offset = get_int(dgram, start_index)
    # Make the size a multiple of 32 bits.
    total_size = size + (-size % _BLOB_DGRAM_PAD)
    end_index = int_offset + size
    if end_index - start_index > len(dgram[start_index:]):
        raise ParseError("Datagram is too short.")
    return dgram[int_offset : int_offset + size], int_offset + total_size


def write_blob(val: bytes) -> bytes:
    """Returns the datagram for the given blob parameter value.

    Raises:
      - BuildError if the value was empty or if its size didn't fit an OSC int.
    """
    if not val:
        raise BuildError("Blob value cannot be empty")
    dgram = write_int(len(val))
    dgram += val
    while len(dgram) % _BLOB_DGRAM_PAD != 0:
        dgram += b"\x00"
    return dgram


def get_date(dgram: bytes, start_index: int) -> tuple[float, int]:
    """Get a 64-bit big-endian fixed-point time tag as a date from the datagram.

    According to the specifications, a date is represented as is:
    "the first 32 bits specify the number of seconds since midnight on
    January 1, 1900, and the last 32 bits specify fractional parts of a second
    to a precision of about 200 picoseconds".

    Args:
      dgram: A datagram packet.
      start_index: An index where the date starts in the datagram.

    Returns:
      A tuple containing the system date and the new end index.
      returns osc_immediately (0) if the corresponding OSC sequence was found.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    # Check for the special case first.
    if dgram[start_index : start_index + _TIMETAG_DGRAM_LEN] == ntp.IMMEDIATELY:
        return IMMEDIATELY, start_index + _TIMETAG_DGRAM_LEN
    if len(dgram[start_index:]) < _TIMETAG_DGRAM_LEN:
        raise ParseError("Datagram is too short")
    timetag, start_index = get_uint64(dgram, start_index)
    seconds = timetag * ntp._NTP_TIMESTAMP_TO_SECONDS
    return ntp.ntp_time_to_system_epoch(seconds), start_index


def write_date(system_time: int | float) -> bytes:
    if system_time == IMMEDIATELY:
        return ntp.IMMEDIATELY

    try:
        return ntp.system_time_to_ntp(system_time)
    except ntp.NtpError as ntpe:
        raise BuildError(ntpe)


def write_rgba(val: bytes) -> bytes:
    """Returns the datagram for the given rgba32 parameter value

    Raises:
      - BuildError if the int could not be converted.
    """
    try:
        return struct.pack(">I", val)
    except struct.error as e:
        raise BuildError(f"Wrong argument value passed: {e}")


def get_rgba(dgram: bytes, start_index: int) -> tuple[bytes, int]:
    """Get an rgba32 integer from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the integer starts in the datagram.

    Returns:
      A tuple containing the integer and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _INT_DGRAM_LEN:
            raise ParseError("Datagram is too short")
        return (struct.unpack(">I", dgram[start_index : start_index + _INT_DGRAM_LEN])[0], start_index + _INT_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")


MIDI_PACKET_LENGTH = 4


def write_midi(val: MidiPacket) -> bytes:
    """Returns the datagram for the given MIDI message parameter value

       A valid MIDI message: (port id, status byte, data1, data2).

    Raises:
      - BuildError if the MIDI message could not be converted.

    """
    if len(val) != MIDI_PACKET_LENGTH:
        raise BuildError("MIDI message length is invalid")
    try:
        value = sum((value & 0xFF) << 8 * (3 - pos) for pos, value in enumerate(val))
        return struct.pack(">I", value)
    except struct.error as e:
        raise BuildError(f"Wrong argument value passed: {e}")


def get_midi(dgram: bytes, start_index: int) -> tuple[MidiPacket, int]:
    """Get a MIDI message (port id, status byte, data1, data2) from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the MIDI message starts in the datagram.

    Returns:
      A tuple containing the MIDI message and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _INT_DGRAM_LEN:
            raise ParseError("Datagram is too short")
        val = struct.unpack(">I", dgram[start_index : start_index + _INT_DGRAM_LEN])[0]
        midi_msg = cast(MidiPacket, tuple((val & 0xFF << 8 * i) >> 8 * i for i in range(3, -1, -1)))
        return (midi_msg, start_index + _INT_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise ParseError(f"Could not parse datagram {e}")



# ruff: noqa: B904,ANN401
from __future__ import annotations

import logging
from typing import Any, Iterator

from fastosc.message import osc_message
from fastosc.message.parsing import osc_types

_BUNDLE_PREFIX = b"#bundle\x00"


class ParseError(Exception):
    """Base exception raised when a datagram parsing error occurs."""


class OscBundle:
    """Bundles elements that should be triggered at the same time.

    An element can be another OscBundle or an OscMessage.
    """

    def __init__(self, dgram: bytes) -> None:
        """Initializes the OscBundle with the given datagram.

        Args:
          dgram: a UDP datagram representing an OscBundle.

        Raises:
          ParseError: if the datagram could not be parsed into an OscBundle.
        """
        # Interesting stuff starts after the initial b"#bundle\x00".
        self._dgram = dgram
        index = len(_BUNDLE_PREFIX)
        try:
            self._timestamp, index = osc_types.get_date(self._dgram, index)
        except osc_types.ParseError as pe:
            raise ParseError(f"Could not get the date from the datagram: {pe}")
        # Get the contents as a list of OscBundle and OscMessage.
        self._contents = self._parse_contents(index)

    def _parse_contents(self, index: int) -> list[OscBundle | osc_message.OscMessage]:
        contents = []  # type: list[OscBundle | osc_message.OscMessage]

        try:
            # An OSC Bundle Element consists of its size and its contents.
            # The size is an int32 representing the number of 8-bit bytes in the
            # contents, and will always be a multiple of 4. The contents are either
            # an OSC Message or an OSC Bundle.
            while self._dgram[index:]:
                # Get the sub content size.
                content_size, index = osc_types.get_int(self._dgram, index)
                # Get the datagram for the sub content.
                content_dgram = self._dgram[index : index + content_size]
                # Increment our position index up to the next possible content.
                index += content_size
                # Parse the content into an OSC message or bundle.
                if OscBundle.dgram_is_bundle(content_dgram):
                    contents.append(OscBundle(content_dgram))
                elif osc_message.OscMessage.dgram_is_message(content_dgram):
                    contents.append(osc_message.OscMessage(content_dgram))
                else:
                    logging.warning(f"Could not identify content type of dgram {content_dgram!r}")
        except (osc_types.ParseError, osc_message.ParseError, IndexError) as e:
            raise ParseError(f"Could not parse a content datagram: {e}")

        return contents

    @staticmethod
    def dgram_is_bundle(dgram: bytes) -> bool:
        """Returns whether this datagram starts like an OSC bundle."""
        return dgram.startswith(_BUNDLE_PREFIX)

    @property
    def timestamp(self) -> float:
        """Returns the timestamp associated with this bundle."""
        return self._timestamp

    @property
    def num_contents(self) -> int:
        """Shortcut for len(*bundle) returning the number of elements."""
        return len(self._contents)

    @property
    def size(self) -> int:
        """Returns the length of the datagram for this bundle."""
        return len(self._dgram)

    @property
    def dgram(self) -> bytes:
        """Returns the datagram from which this bundle was built."""
        return self._dgram

    def content(self, index: int) -> Any:
        """Returns the bundle's content 0-indexed."""
        return self._contents[index]

    def __iter__(self) -> Iterator[Any]:
        """Returns an iterator over the bundle's content."""
        return iter(self._contents)



from __future__ import annotations

from fastosc.message.arg_value import ArgValue
from fastosc.message.osc_message_builder import BuildError, OscMessageBuilder


def convert_message(*, address: str, params: list[ArgValue]) -> bytes:
    msg_builder = OscMessageBuilder(address)
    for param in params or []:
        msg_builder.add_arg(param)
    try:
        msg = msg_builder.build()
        return msg.dgram
    except BuildError as e:
        raise e



from __future__ import annotations

import logging
import socket

from fastosc.dispatcher import Dispatcher
from fastosc.message.osc_bundle import OscBundle
from fastosc.message.osc_message import OscMessage
from fastosc.server.dispatcher_server import OSCDispatcherServer


class OSCUDPServerBase(OSCDispatcherServer):
    def __init__(
        self,
        *,
        dispatcher: Dispatcher,
        logger: logging.Logger,
        local_addr: tuple[str, int],
    ) -> None:
        super().__init__(logger=logger, local_addr=local_addr, dispatcher=dispatcher)
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self._socket.bind(self._local_addr)

    def _send_bytes(self, data: bytes, remote_addr: tuple[str, int]) -> None:
        self._socket.sendto(data, remote_addr)

    def _parse_datagram(self, *, data: bytes, remote_addr: tuple[str, int]) -> None:
        if OscMessage.dgram_is_message(data):
            self._dispatcher.process_message(message=OscMessage(data), remote_addr=remote_addr)
        elif OscBundle.dgram_is_bundle(data):
            self._dispatcher.process_bundle(bundle=OscBundle(data), remote_addr=remote_addr)
        else:
            logging.debug(f"unknown osc message: {data} from {remote_addr}")  # type: ignore[str-bytes-safe]



from __future__ import annotations

import errno
import logging
import traceback

from fastosc.dispatcher import Dispatcher
from fastosc.server.udp.udp_server_base import OSCUDPServerBase


class OSCUDPPullServer(OSCUDPServerBase):
    """
    OSC Udp server that stores datagrams and has a process() method to pull in stored datagrams.
    This is more efficient in certain environments such as Ableton live embedded python which
    does not support multithreading well (i.e it gives latencies in 100+ms and cannot handle high rates of inbound
    messages without spiking latency.
    """

    def __init__(self, *, dispatcher: Dispatcher, logger: logging.Logger, local_addr: tuple[str, int]) -> None:
        super().__init__(logger=logger, dispatcher=dispatcher, local_addr=local_addr)
        self._socket.setblocking(False)  # noqa: FBT003

    def process(self) -> None:
        """
        Synchronously process all data queued on the OSC socket.
        """
        try:
            while True:
                data, remote_addr = self._socket.recvfrom(65536)
                self._parse_datagram(data=data, remote_addr=remote_addr)

        except OSError as e:
            if e.errno == errno.ECONNRESET:
                # --------------------------------------------------------------------------------
                # This benign error seems to occur on startup on Windows
                # --------------------------------------------------------------------------------
                self._logger.warning(f"Non-fatal socket error: {traceback.format_exc()}")
            elif e.errno in (errno.EAGAIN, errno.EWOULDBLOCK):
                # --------------------------------------------------------------------------------
                # Another benign networking error, throw when no data is received
                # on a call to recvfrom() on a non-blocking socket
                # --------------------------------------------------------------------------------
                pass
            else:
                # --------------------------------------------------------------------------------
                # Something more serious has happened
                # --------------------------------------------------------------------------------
                self._logger.error(f"Socket error: {traceback.format_exc()}")

        except Exception as e:  # noqa
            self._logger.error(f"Error handling OSC message: {e}")
            self._logger.warning(f"{traceback.format_exc()}")

    def shutdown(self) -> None:
        """
        Shutdown the server network sockets.
        """
        self._socket.close()



"""Representation of an OSC message in a pythonesque way."""
# ruff: noqa: B904,PLR0912,PLR0915,PLR2004

from __future__ import annotations

import logging
from typing import Iterator

from fastosc.message.arg_value import ArgValue
from fastosc.message.parsing import osc_types


class ParseError(Exception):
    """Base exception raised when a datagram parsing error occurs."""


class OscMessage:
    """Representation of a parsed datagram representing an OSC message.

    An OSC message consists of an OSC Address Pattern followed by an OSC
    Type Tag String followed by zero or more OSC Arguments.
    """

    def __init__(self, dgram: bytes) -> None:
        self._dgram = dgram
        self._parameters: list[ArgValue] = []  #
        self._parse_datagram()

    def _parse_datagram(self) -> None:
        try:
            self._address_regexp, index = osc_types.get_string(self._dgram, 0)
            if not self._dgram[index:]:
                # No params is legit, just return now.
                return

            # Get the parameters types.
            type_tag, index = osc_types.get_string(self._dgram, index)
            if type_tag.startswith(","):
                type_tag = type_tag[1:]

            params: list[ArgValue] = []
            param_stack = [params]
            # Parse each parameter given its type.
            for param in type_tag:
                val = NotImplemented  # Any
                if param == "i":  # Integer.
                    val, index = osc_types.get_int(self._dgram, index)  # type: ignore
                elif param == "h":  # Int64.
                    val, index = osc_types.get_int64(self._dgram, index)  # type: ignore
                elif param == "f":  # Float.
                    val, index = osc_types.get_float(self._dgram, index)  # type: ignore
                elif param == "d":  # Double.
                    val, index = osc_types.get_double(self._dgram, index)  # type: ignore
                elif param == "s":  # String.
                    val, index = osc_types.get_string(self._dgram, index)  # type: ignore
                elif param == "b":  # Blob.
                    val, index = osc_types.get_blob(self._dgram, index)  # type: ignore
                elif param == "r":  # RGBA.
                    val, index = osc_types.get_rgba(self._dgram, index)  # type: ignore
                elif param == "m":  # MIDI.
                    val, index = osc_types.get_midi(self._dgram, index)  # type: ignore
                elif param == "t":  # osc time tag:
                    val, index = osc_types.get_timetag(self._dgram, index)  # type: ignore
                elif param == "T":  # True.
                    val = True  # type: ignore
                elif param == "F":  # False.
                    val = False  # type: ignore
                elif param == "N":  # Nil.
                    val = None  # type: ignore
                elif param == "[":  # Array start.
                    array: list[ArgValue] = []
                    param_stack[-1].append(array)
                    param_stack.append(array)
                elif param == "]":  # Array stop.
                    if len(param_stack) < 2:
                        raise ParseError(f"Unexpected closing bracket in type tag: {type_tag}")
                    param_stack.pop()
                # TODO: Support more exotic types as described in the specification.
                else:
                    logging.warning(f"Unhandled parameter type: {param}")
                    continue
                if param not in "[]":
                    param_stack[-1].append(val)
            if len(param_stack) != 1:
                raise ParseError(f"Missing closing bracket in type tag: {type_tag}")
            self._parameters = params
        except osc_types.ParseError as pe:
            raise ParseError("Found incorrect datagram, ignoring it", pe)

    @property
    def address(self) -> str:
        """Returns the OSC address regular expression."""
        return self._address_regexp

    @staticmethod
    def dgram_is_message(dgram: bytes) -> bool:
        """Returns whether this datagram starts as an OSC message."""
        return dgram.startswith(b"/")

    @property
    def size(self) -> int:
        """Returns the length of the datagram for this message."""
        return len(self._dgram)

    @property
    def dgram(self) -> bytes:
        """Returns the datagram from which this message was built."""
        return self._dgram

    @property
    def params(self) -> list[ArgValue]:
        """Convenience method for list(self) to get the list of parameters."""
        return list(self)

    def __iter__(self) -> Iterator[ArgValue]:
        """Returns an iterator over the parameters of this message."""
        return iter(self._parameters)
